%{

  /*
   * Simplified flex scanner for C++ code
   *
   * Stephen Edwards
   */

#include "cceInt.h"
#include "cceGram.h"
#include <stdlib.h>
#include <string.h>

extern YYLTYPE ccelloc;


/*
 * Line number of current token
 */
int ccelineno;

#define SAVECCETEXT strcpy(malloc(cceleng+1),ccetext)
#define SAVEPARTIALCCETEXT(x) strcpy(malloc(cceleng+1-x),ccetext+x)

/* Uncomment to print each token as it is recognized */
/* #define YY_USER_ACTION printf("%d: %s\n",yylineno,yytext); */

#define YY_USER_ACTION \
  ccelloc.first_line = ccelineno;

#define ECHO \
  printf("scanner didn't recognize `%s' (%d) at line %d\n", yytext, yytext[0], ccelineno)

%}

  /* Output to cceScan.c */
%option outfile="cceScan.c"

  /* Prefix everything with "cce" instead of "yy" */
%option prefix="cce"

%x comment

identifier	[a-zA-Z_][0-9a-zA-Z_]*

exponentPart		[eE][-+]?[0-9]+
fractionalConstant	([0-9]*"."[0-9]+)|([0-9]+".")
floatingConstant	(({fractionalConstant}{exponentPart}?)|([0-9]+{exponentPart}))[FfLl]?

optIntegerSuffix	([uU]?[lL]?)|([lL][uU])
decimalConstant		[1-9][0-9]*{optIntegerSuffix}
octalConstant		"0"[0-7]*{optIntegerSuffix}
hexConstant		"0"[xX][0-9a-fA-F]+{optIntegerSuffix}

number		-?({floatingConstant}|{decimalConstant}|{octalConstant}|{hexConstant})

whitespace	[ \t\f]

cppLine		^#(.*\\[ \t]*\n)*.*\n

quotedString	(\"([^"\n]|\\.)*\")|'(.|(\\.))'

oldstyleComment	"/*"([^*]|[*]*[^*/])*[*]*"*/"
commentLine	"//".*\n
newstyleComment {commentLine}(({whitespace})*{commentLine})*

op0		"="|"|"|"^"|"<"|">"|"+"|"-"|"!"|"~"|"."|"/"|"%"
op1		"->"|"++"|"--"|".*"|"->*"|"<<"|">>"|"<="|">="|"=="|"!="|"&&"
op2		"||"|"*="|"/="|"%="|"+="|"-="|"<<="|">>="|"&="|"^="|"|="|"::"
op3		"new"|"delete"
operator	{op0}|{op1}|{op2}|{op3}

%%

\n		{ ccelineno++; }
{whitespace}+	{}

"/""*"+		{ BEGIN(comment); }
<comment>([^@*]|"*"*[^@*/])* {

		  char * p = ccetext;
		  char c;

	          /*
                   * Add the number of newlines in the string to the total
		   */

		  while ( (c = *p++) ) {
                    if ( c == '\n' ) {
		      ccelineno++;
		    }
	   	  }

		  ccelval.string = SAVECCETEXT;
		  return COMMENT;
		}
<comment>"@"{identifier} {
		  ccelval.string = SAVEPARTIALCCETEXT(1);
		  return COMMENTFIELD;
		}
<comment>"@" {}
<comment>"*"+"/" { BEGIN(INITIAL); }

{newstyleComment} {

  /*
   * We have a C++-style comment, possibly multi-line:
   * Strip out the whitespace followed by a // and push it back to make
   * it look like an old-style comment
   */
  
  char * s;
  char * d;
  s = d = ccetext;

  /*
   * Skip characters until after //, then swallow as many slashes as possible,
   * then copy characters up to and including a newline
   */

  do {
    while ( *s++ != '/' ) {}
    while ( *s == '/' ) { s++; }
    while ( (*d++ = *s++) != '\n' ) {}
  } while ( s < ccetext + cceleng );

  /*
   * Push back the string with a trailing comment-end pair.
   * This should not overwrite the ccetext buffer, since we removed at least
   * two characters--the first two slashes
   */

  unput('\n');
  unput('/');
  unput('*');
  --d;
  while ( --d >= ccetext ) {
    unput(*d);
  }
  
  BEGIN(comment);
}

class		{ return CLASS; }
struct		{ return STRUCT; }
union		{ return UNION; }
enum		{ return ENUM; }
friend		{ return FRIEND; }
operator	{ return OPERATOR; }
typedef		{ return TYPEDEF; }
extern		{ return EXTERN_T; }

private		{ ccelval.i = CCE_PRIVATE;	return ACCESS; }
protected	{ ccelval.i = CCE_PROTECTED;	return ACCESS; }
public		{ ccelval.i = CCE_PUBLIC;	return ACCESS; }

const		{ ccelval.i = CCE_CONST;	return QUALIFIER; }
static		{ ccelval.i = CCE_STATIC;	return QUALIFIER; }
virtual		{ ccelval.i = CCE_VIRTUAL;	return QUALIFIER; }
inline		{ ccelval.i = CCE_INLINE;	return QUALIFIER; }
unsigned	{ ccelval.i = CCE_UNSIGNED;	return QUALIFIER; }
short		{ ccelval.i = CCE_SHORT;	return QUALIFIER; }
long		{ ccelval.i = CCE_LONG;		return QUALIFIER; }

"::"		{ return DOUBLECOLON; }
[,{}();:*&~\[\]=0] { return ccetext[0]; }
{operator}	{ ccelval.string = SAVECCETEXT; return OP; }

{identifier}	{ ccelval.string = SAVECCETEXT; return IDENTIFIER; }

{cppLine}	{
 		  char * p = ccetext;
		  char c;

	          /*
                   * Add the number of newlines in the string to the total
		   */

		  while ( (c = *p++) ) {
                    if ( c == '\n' ) {
		      ccelineno++;
		    }
	   	  }
		}

{quotedString}  { ccelval.string = SAVECCETEXT; return TOKEN; }
{number}|.	{ ccelval.string = SAVECCETEXT; return TOKEN; }

%%

ccewrap()
{
  return 1;
}

/* Local Variables: */
/* mode: fundamental */
/* End: */

