#
# cceSchema.tcl
#
# A Tcl script that generates C code to generate the built-in objects
# for the cce system.  Invoked by the Makefile
#

set cce_slotTypeConstants(int) CCE_INTEGER
set cce_slotTypeConstants(string) CCE_STRING
set cce_slotTypeConstants(object) CCE_OBJECT

set cce_slotTypeReaders(int) Cce_ObjectSlotReadInt
set cce_slotTypeReaders(string) Cce_ObjectSlotReadString
set cce_slotTypeReaders(object) Cce_ObjectSlotReadObject

# C Code for creating the objects
set createObjectCode ""

# C Code defining the variables for each class
set variableDefinitions ""

# C Code defining the external variables for each class
set externalVariables ""

# Constructors, slot accessor macros
set headerMacros ""

# Formatted HTML documentation
set classDoc ""

proc cce_class {className classComment slots} {
    global createObjectCode
    global variableDefinitions
    global externalVariables
    global headerMacros
    global classDoc
    global cce_slotTypeConstants
    global cce_slotTypeReaders

    # Internal C variable

    set cvar "cce_[string tolower [string range $className 0 0]][string range $className 1 end]Class"

    append variableDefinitions "Cce_Object $cvar;\n";

    append externalVariables "EXTERN Cce_Object $cvar;\n";

    append createObjectCode "  $cvar = Cce_ClassNew( \"$className\" );\n"

    append headerMacros "\n/*\n * Macros for the $className class\n */\n"
    append headerMacros "\#define Cce_${className}New() Cce_ObjectNew( $cvar )\n\n"

    append classDoc "<H3>Class <TT>$className</TT></H3>\n\n"
    append classDoc "$classComment <P>\n\n"
    append classDoc "<TABLE>\n";

    # Starting slot number is 3, past the class, id, and next fields
    set slotNumber 3
    set index 0
    while { [set slotName [lindex $slots $index]] != "" } {
	incr index
	set slotType [lindex $slots $index]
	incr index
	set slotComment [lindex $slots $index]
	incr index

	append createObjectCode "  Cce_ClassAddSlot( $cvar, \"$slotName\", $cce_slotTypeConstants($slotType) );\n"

	append headerMacros "\#define Cce_${className}Read[string toupper [string range $slotName 0 0]][string range $slotName 1 end](o) $cce_slotTypeReaders($slotType)(o, $slotNumber)\n"

	append headerMacros "\#define Cce_${className}Write[string toupper [string range $slotName 0 0]][string range $slotName 1 end](o,v) Cce_ObjectSlotWrite(o, $slotNumber, v)\n"

	append classDoc "<TR> <TD><TT>$slotName</TT></TD>\n"
	append classDoc "     <TD><I>$slotType</I></TD>\n"
	append classDoc "     <TD>$slotComment</TD></TR>\n"

	incr slotNumber
    }

    append classDoc "</TABLE>\n\n"
}

######################################################################
#
# Definitions of the built-in objects
#

cce_class CMethod {A C++ class method} {
    line int {declaration line number}
    file string {declaration file name}
    comment object {documenting comment (CComment)}
    access int {access type: 0 = private, 1 = protected, 2 = public.  Valid only for declarations}

    name string {name of the method}
    className string {name of this method's class}

    type object {return type of the method (CType)}
    arguments object {list of arguments (CArgument)}
    qualifiers int {bit fields:  2 = const, 256 = pure }
}

cce_class COperator {A C++ operator} {
    line int {declaration line number}
    file string {declaration file name}
    comment object {documenting comment (CComment)}
    access int {access type: 0 = private, 1 = protected, 2 = public.  Valid only for declarations}

    operator string {operator or "cast"}
    className string {name of this operator's class}

    type object {return type of operator}
    arguments object {list of arguments (CArgument)}
    qualifiers int {bit fields: 2 = const, 256 = pure }
}

cce_class CVariable {A variable definition} {
    line int {declaration line number}
    file string {declaration file name}
    comment object {documenting comment (CComment)}
    access int {access type: 0 = private, 1 = protected, 2 = public.  Valid only for declarations}

    name string {}
    className string {name of this variable's class}

    type object {(CType)}
}

cce_class CClass {A C++ class definition} {
    line int {declaration line number}
    file string {declaration file name}
    comment object {documenting comment (CComment)}
    access int {access type: 0 = private, 1 = protected, 2 = public}

    name string {name of the class}

    inherits object {list of inherited classes (CClassInherits)}
    members object {list of members}
}

cce_class CClassInherits {Inheritance information for a C++ class} {
    name string {name of inherited class}
    access int {access type: 0 = private, 1 = protected, 2 = public}    
}

cce_class CType {A C++ variable, method, etc. type} {
    name string {the type (int, class name, etc)}
    qualifiers int {bit fields: 1 = virtual, 2 = const, 4 = static, 8 = inline, 16 = unsigned, 32 = short, 64 = long, 128 = array, 256 = pure }
    indirect int { -1 = reference, 0 = direct, 1 = pointer to, 2 = **, etc. }
}

cce_class CArgument {An argument to a method, function, etc.} {
    name string {}
    type object {(CType)}
    comment object {documenting comment (CComment)}
    default string {default value}
}

cce_class CComment {A field in a comment} {
    field string {Name of the field, if any}
    value string {Value of the field, if any}
}

#
# End of object definitions
#
######################################################################

set f [open cceObjectVariables.include w]
puts $f $variableDefinitions
close $f

set f [open cceObjectHeader.include w]
puts $f $externalVariables
puts $f $headerMacros
close $f

set f [open cceObjectCreators.include w]
puts $f $createObjectCode
close $f

set f [open cceObjectDoc.html w]
puts $f "<HTML><HEAD><TITLE>CCE Built-in Objects</TITLE></HEAD>
<BODY>
<H1>CCE Built-in Objects</H1>

<HR>

$classDoc

<HR>
Generated automatically by cceSchema <BR>
Last updated [exec date +%y%m%d]
</BODY>
</HTML>"
close $f


